import base64
import unittest

from youtube_simple_scraper.continuation_token import CommentsContinuationTokenBuilder
from youtube_simple_scraper.proto import continuation


class TestContinuationToken(unittest.TestCase):

    def test_parse_continuation_for_video_list(self):
        expected = "CggIgAQVF7fROBIFCIcgGAASBQiIIBgAEgUInSAYARIFCIkgGAASBQioIBgAGAAiDQoLCNXK8rIGEKj4iS8"
        p = CommentsContinuationTokenBuilder.build_params_str_for_video_list(1717347669)
        self.assertEqual(p, expected)
        serializedb64 = CommentsContinuationTokenBuilder.build_video_list_token("bkRcuK2i7oM", 60, 1717347669)
        expected = ("Eg0SC2JrUmN1SzJpN29NGAYyjgEKZWdldF9uZXdlc3RfZmlyc3QtLUNnZ0lnQVFW"
                    "RjdmUk9CSUZDSWNnR0FBU0JRaUlJQmdBRWdVSW5TQVlBUklGQ0lrZ0dBQVNCUWlvSUJnQUdBQWlEUW9MQ0"
                    "5YSzhySUdFS2o0aVM4IhEiC2JrUmN1SzJpN29NMAF4ASg8QhBjb21tZW50cy1zZWN0aW9u")
        self.assertEqual(serializedb64, expected)

    def test_parse_continuation_for_short_list(self):
        expected = ("4qmFsgJ-Eh5GRWNvbW1lbnRfc2hvcnRzX3dlYl90b3BfbGV2ZWwaXHFnTkNJ"
                    "aFFpQzFBNWIyOHRja0psTjA5Vk1BQjRBc2dCQURBQlFpaHphRzl5ZEhNdFpX"
                    "NW5ZV2RsYldWdWRDMXdZVzVsYkMxamIyMXRaVzUwY3kxelpXTjBhVzl1")
        p = CommentsContinuationTokenBuilder.build_start_short_comment_cont_token("P9oo-rBe7OU")
        self.assertEqual(p, expected)



import datetime
import json
from typing import Tuple, List

import dateparser
import requests

from youtube_simple_scraper.continuation_token import CommentsContinuationTokenBuilder
from youtube_simple_scraper.counters import comment_counter_to_int

USER_AGENT = ("Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
              "AppleWebKit/537.36 (KHTML, like Gecko) "
              "Chrome/125.0.0.0 Safari/537.36,gzip(gfe)")


def build_list_videos_request_headers(channel_id: str) -> dict:
    return {
        "accept": "*/*",
        "accept-language": "en",
        "content-type": "application/json",
        "origin": "https://www.youtube.com",
        "priority": "u=1, i",
        "referer": f"https://www.youtube.com/@{channel_id}",
        "sec-ch-ua": '"Google Chrome";v="125", "Chromium";v="125", "Not.A/Brand";v="24"',
        "sec-ch-ua-arch": '"arm"',
        "sec-ch-ua-bitness": '"64"',
        "sec-ch-ua-full-version": '"125.0.6422.78"',
        "sec-ch-ua-full-version-list": '"Google Chrome";v="125.0.6422.78", '
                                       '"Chromium";v="125.0.6422.78", '
                                       '"Not.A/Brand";v="24.0.0.0"',
        "sec-ch-ua-mobile": '?0',
        "sec-ch-ua-model": '"',
        "sec-ch-ua-platform": '"macOS"',
        "sec-ch-ua-platform-version": '"14.4.1"',
        "sec-ch-ua-wow64": '?0',
        "sec-fetch-dest": 'empty',
        "sec-fetch-mode": 'same-origin',
        "sec-fetch-site": "same-origin",
        "user-agent": USER_AGENT,
        "x-youtube-bootstrap-logged-in": 'false',
        "x-youtube-client-name": '1',
        "x-youtube-client-version": '2.20240530.02.00'
    }


def build_list_videos_request_payload(channel_id: str, browse_id: str, continuation_token: str) -> dict:
    base = {
        "context": {
            "client": {
                "deviceMake": "Apple",
                "userAgent": USER_AGENT,
                "clientName": "WEB",
                "clientVersion": "2.20240530.02.00",
                "osName": "Macintosh",
                "osVersion": "10_15_7",
                "originalUrl": f"https://www.youtube.com/@{channel_id}",
                "screenPixelDensity": 2,
                "platform": "DESKTOP",
                "clientFormFactor": "UNKNOWN_FORM_FACTOR",
                "screenDensityFloat": 2,
                "userInterfaceTheme": "USER_INTERFACE_THEME_DARK",
                "browserName": "Chrome",
                "browserVersion": "125.0.0.0",
                "acceptHeader": "text/html,application/xhtml+xml,application/xml;q=0.9,"
                                "image/avif,image/webp,image/apng,*/*;q=0.8,"
                                "application/signed-exchange;v=b3;q=0.7",
                "screenWidthPoints": 1526,
                "screenHeightPoints": 1546,
                "utcOffsetMinutes": -240,
            },
            "user": {
                "lockedSafetyMode": True
            },
            "request": {
                "useSsl": True,
                "internalExperimentFlags": [],
                "consistencyTokenJars": []
            },
        },
        "browseId": browse_id,
    }
    if continuation_token:
        base["continuation"] = continuation_token
    return base


def find_channel_basic_info(channel_id: str) -> dict:
    url = f"https://www.youtube.com/@{channel_id}/videos"
    headers = {
        'accept': 'application/json',
        'accept-language': 'en,en-US;q=0.9,en;q=0.8,pt;q=0.7',
        'cache-control': 'max-age=0',
        'service-worker-navigation-preload': 'true',
        'upgrade-insecure-requests': '1',
        'user-agent': USER_AGENT,
    }
    response = requests.request("GET", url, headers=headers)
    start_tag = "ytInitialData = "
    return _extract_channel_basic_info_from_html(response, start_tag)


def get_comments_from_api(video_id: str, token: str = "") -> Tuple[List[dict], str]:
    if token:
        continuation_token = token
    else:
        continuation_token = CommentsContinuationTokenBuilder.build_video_list_token(video_id, 0)
    url = "https://www.youtube.com/youtubei/v1/next?prettyPrint=false"
    payload = json.dumps({
        "context": {
            "client": {
                "deviceModel": "",
                "userAgent": USER_AGENT,
                "clientName": "WEB",
                "clientVersion": "2.20240530.02.00",
                "screenPixelDensity": 2,
                "platform": "DESKTOP",
                "clientFormFactor": "UNKNOWN_FORM_FACTOR",
                "screenDensityFloat": 2,
                "userInterfaceTheme": "USER_INTERFACE_THEME_DARK",
                "timeZone": "America/Santiago",
                "browserName": "Chrome",
                "browserVersion": "125.0.0.0",
                "acceptHeader": "text/html,application/xhtml+xml,application/xml;q=0.9,"
                                "image/avif,image/webp,image/apng,*/*;q=0.8,"
                                "application/signed-exchange;v=b3;q=0.7",
                "screenWidthPoints": 1501,
                "screenHeightPoints": 1461,
                "utcOffsetMinutes": -240,
                "connectionType": "CONN_CELLULAR_4G",
                "memoryTotalKbytes": "8000000"
            },
            "user": {
                "lockedSafetyMode": False
            },
            "request": {
                "useSsl": True,
                "internalExperimentFlags": [],
                "consistencyTokenJars": []
            }
        },
        "continuation": continuation_token
    })
    headers = {
        'accept': '*/*',
        'accept-language': 'en',
        'content-type': 'application/json',
        'origin': 'https://www.youtube.com',
        'x-youtube-client-version': '2.20240530.02.00',
    }
    response = requests.request("POST", url, headers=headers, data=payload)
    comments, token, is_valid = _extract_raw_comments(response)
    if not is_valid:
        return get_comments_from_api(video_id, token)
    return comments, token


def _extract_raw_comments(response) -> Tuple[List[dict], str, bool]:
    r = response.json()
    comments = []
    continuation_token = ""
    if "onResponseReceivedEndpoints" in r:
        for ep in r["onResponseReceivedEndpoints"]:
            if "appendContinuationItemsAction" not in ep:
                continue
            for ci in ep["appendContinuationItemsAction"]["continuationItems"]:
                if "continuationItemRenderer" in ci:
                    continuation_token = ci["continuationItemRenderer"]["continuationEndpoint"][
                        "continuationCommand"]["token"]
    if "frameworkUpdates" in r:
        for ep in r["frameworkUpdates"]["entityBatchUpdate"]["mutations"]:
            if "commentEntityPayload" in ep["payload"]:
                prop = ep["payload"]["commentEntityPayload"]["properties"]
                comment_id = prop["commentId"]
                comment_text = prop["content"]["content"]
                author = prop["authorButtonA11y"]
                published_time = dateparser.parse(prop["publishedTime"].replace("(edited)", ""), languages=["en"])
                if published_time is None:
                    published_time = datetime.datetime.now()
                likes = 0
                toolbar = ep["payload"]["commentEntityPayload"]["toolbar"]
                if "likeCountA11y" in toolbar:
                    likes = comment_counter_to_int(toolbar["likeCountA11y"])
                replies_count = 0
                if "replyCountA11y" in toolbar:
                    replies_count = comment_counter_to_int(toolbar["replyCountA11y"])
                comments.append({
                    "id": comment_id,
                    "author": author,
                    "published_time": published_time,
                    "likes": likes,
                    "comment_text": comment_text,
                    "replies_count": replies_count
                })
        return comments, continuation_token, True
    if 'responseContext' in r:
        return [], continuation_token, True
    return [], continuation_token, False


def _extract_channel_basic_info_from_html(response, start_tag) -> dict:
    start_idx = response.text.find(start_tag)
    if start_idx == -1:
        return {}
    start_idx += len(start_tag)
    end_tag = "</script>"
    end_idx = response.text.find(end_tag, start_idx)
    if end_idx == -1:
        return {}
    json_str = response.text[start_idx:end_idx].strip()
    json_str = json_str[:json_str.rfind(";")]
    return json.loads(json_str)



from typing import List

from youtube_simple_scraper.entities import ListVideoStopCondition, ListCommentStopCondition, Video, VideoComment


class ListVideoMaxPagesStopCondition(ListVideoStopCondition):
    def __init__(self, max_pages: int):
        self._max_pages = max_pages

    def should_stop(self, videos: List[Video], page: int) -> bool:
        return page >= self._max_pages


class ListVideoNeverStopCondition(ListVideoStopCondition):
    def should_stop(self, videos: List[Video], page: int) -> bool:
        return False


class ListCommentMaxPagesStopCondition(ListCommentStopCondition):
    def __init__(self, max_pages: int):
        self._max_pages = max_pages

    def should_stop(self, video: Video, comments: List[VideoComment], page: int) -> bool:
        return page >= self._max_pages


class ListCommentNeverStopCondition(ListCommentStopCondition):
    def should_stop(self, video: Video, comments: List[VideoComment], page: int) -> bool:
        return False



import re

import regex


def comment_counter_to_int(c: str) -> int:
    """
    Convert comment counter string to integer.

    :param c: Comment counter string.
    :return: Converted integer value.

    Example:
    >>> comment_counter_to_int("12 Views")
    12
    >>> comment_counter_to_int("1K Views")
    1000
    >>> comment_counter_to_int("1.2K Views")
    1200
    >>> comment_counter_to_int("12.3K Views")
    12300
    >>> comment_counter_to_int("123.4K Views")
    123400
    >>> comment_counter_to_int("123K Views")
    123000
    >>> comment_counter_to_int("1.23K Views")
    1230
    >>> comment_counter_to_int("1.234M Views")
    1234000
    >>> comment_counter_to_int("12M Likes")
    12000000
    >>> comment_counter_to_int("12.34M Likes")
    12340000
    """
    c = c.upper()
    r = re.compile(r'(?P<number>[\d.,]+)\s*(?P<mult>[MK])?.*')
    match = r.match(c)
    if not match:
        return 0
    number = match.group('number')
    mult_char = match.group('mult')
    mult = 1
    if mult_char == 'K':
        mult = 1000
    elif mult_char == 'M':
        mult = 1000000
    else:
        number = number.replace(',', '').replace('.', '')
    return int(float(number) * mult)



import base64
import datetime
from typing import Optional

from youtube_simple_scraper.proto import continuation


class CommentsContinuationTokenBuilder:

    @classmethod
    def build_video_list_token(cls, video_id: str, offset: int, timestamp: Optional[int] = 0) -> str:
        if offset > 0:
            if timestamp == 0:
                timestamp = int(datetime.datetime.now(tz=datetime.timezone.utc).timestamp())
            sub_message = cls.build_params_str_for_video_list(timestamp)
            msg = f"get_newest_first--{sub_message}"
        else:
            msg = ""
        token = continuation.ContinuationToken(
            video_id=continuation.VideoIdSubMessage(video_id=video_id),
            field_3=6,
            sub_messsage_6=continuation.ParamsSubMessage(
                string_1=msg,
                sub_messsage_4=continuation.ParamsSubSubMessage(
                    video_id=video_id,
                    int_6=1,
                    int_15=1
                ),
                offset=offset,
                string_8="comments-section"
            )
        )
        serialized_message = token.SerializeToString()
        return base64.urlsafe_b64encode(serialized_message).decode("utf-8").replace("=", "%3D")

    @staticmethod
    def build_params_str_for_video_list(timestamp: int):
        params_message = continuation.ParamsMessage()
        params_message.field_1 = continuation.Field1SubMessage(f1=512, f2=953267991)

        params_message.field_array_2 = [
            continuation.TwinIntMessage(int_1=4103, int_3=0),
            continuation.TwinIntMessage(int_1=4104, int_3=0),
            continuation.TwinIntMessage(int_1=4125, int_3=1),
            continuation.TwinIntMessage(int_1=4105, int_3=0),
            continuation.TwinIntMessage(int_1=4136, int_3=0),
        ]

        params_message.int_3 = 0

        field_4_sub_message = continuation.Field4SubMessage(sub_msg=continuation.TwinInt12Message(
            timestamp=timestamp,
            int_2=98729000
        ))

        params_message.sub_messsage_4 = field_4_sub_message
        b = params_message.SerializeToString()
        params_b64 = base64.urlsafe_b64encode(b).decode("utf-8").replace("=", "")
        return params_b64

    @classmethod
    def build_start_short_comment_cont_token(cls, short_id: str) -> str:
        p = continuation.ShortsRequestPagination(
            sub_message_1=continuation.ShortsRequestPaginationSub(
                sub_message_4=continuation.ShortsRequestPaginationSubSub(
                    video_id=short_id,
                    uint_6=0,
                    uint_15=2,
                    uint_25=0
                ),
                uint_6=1,
                string_8="shorts-engagement-panel-comments-section".encode("utf-8")
            )
        )
        pagination_b64 = base64.urlsafe_b64encode(p.SerializeToString()).decode("utf-8").replace("=", "")
        token = continuation.ShortContinuationToken(
            params=continuation.ShortsRequestSub(
                sort_type=b"FEcomment_shorts_web_top_level",
                pagination=pagination_b64.encode("utf-8")
            )
        )
        serialized_message = token.SerializeToString()
        return base64.urlsafe_b64encode(serialized_message).decode("utf-8").replace("=", "%3D")



import logging
import sys


def build_default_logger() -> logging.Logger:
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    # Cree un manejador para enviar los mensajes al stdout
    handler = logging.StreamHandler(sys.stdout)

    # Puede establecer el nivel del manejador si lo desea
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return logger



import abc
import datetime
from dataclasses import dataclass
from enum import StrEnum
from typing import List, Optional

from pydantic import BaseModel


class ReactionType(StrEnum):
    LIKE = "LIKE"
    DISLIKE = "DISLIKE"


class Reaction(BaseModel):
    type: ReactionType
    count: int


class VideoComment(BaseModel):
    id: str
    text: str
    user: str
    date: datetime.datetime
    likes: int


class VideoType(StrEnum):
    VIDEO = "VIDEO"
    SHORT = "SHORT"


class Video(BaseModel):
    id: str
    title: str
    view_count: int
    thumbnail_url: str
    video_type: VideoType
    description: Optional[str] = ""
    date: Optional[datetime.datetime] = None
    like_count: Optional[int] = 0
    comment_count: Optional[int] = 0
    comments: List[VideoComment] = []

    @property
    def url(self):
        if self.video_type == VideoType.SHORT:
            return f'https://www.youtube.com/shorts/{self.id}'
        return f'https://www.youtube.com/watch?v={self.id}'


class Channel(BaseModel):
    id: str
    name: str
    target_id: str
    title: str
    description: str
    subscriber_count: int
    video_count: int
    url: str
    videos: List[Video]
    shorts: List[Video]


class ListVideoStopCondition(abc.ABC):
    @abc.abstractmethod
    def should_stop(self, videos: List[Video], page: int) -> bool:
        raise NotImplementedError()


class ListCommentStopCondition(abc.ABC):
    @abc.abstractmethod
    def should_stop(self, video: Video, comments: List[VideoComment], page: int) -> bool:
        raise NotImplementedError()


@dataclass
class GetChannelOptions:
    list_video_stop_conditions: List[ListVideoStopCondition]
    list_video_comment_stop_conditions: List[ListCommentStopCondition]
    list_short_stop_conditions: List[ListVideoStopCondition]
    list_short_comment_stop_conditions: List[ListCommentStopCondition]


class VideoListRepository(abc.ABC):
    @abc.abstractmethod
    def get_channel(self, channel_id: str, opts: GetChannelOptions) -> List[Video]:
        raise NotImplementedError()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: continuation.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


@dataclass
class ContinuationToken(betterproto.Message):
    video_id: "VideoIdSubMessage" = betterproto.message_field(2)
    field_3: int = betterproto.uint32_field(3)
    sub_messsage_6: "ParamsSubMessage" = betterproto.message_field(6)


@dataclass
class VideoIdSubMessage(betterproto.Message):
    video_id: str = betterproto.string_field(2)


@dataclass
class ParamsSubMessage(betterproto.Message):
    string_1: str = betterproto.string_field(1)
    sub_messsage_4: "ParamsSubSubMessage" = betterproto.message_field(4)
    offset: int = betterproto.uint32_field(5)
    string_8: str = betterproto.string_field(8)


@dataclass
class ParamsSubSubMessage(betterproto.Message):
    video_id: str = betterproto.string_field(4)
    int_6: int = betterproto.uint32_field(6)
    int_15: int = betterproto.uint32_field(15)


@dataclass
class Field1SubMessage(betterproto.Message):
    f1: int = betterproto.uint32_field(1)
    f2: float = betterproto.fixed32_field(2)


@dataclass
class TwinIntMessage(betterproto.Message):
    int_1: int = betterproto.uint32_field(1)
    int_3: int = betterproto.int32_field(3, group="int_2")


@dataclass
class TwinInt12Message(betterproto.Message):
    timestamp: int = betterproto.uint32_field(1)
    int_2: int = betterproto.uint32_field(2)


@dataclass
class Field4SubMessage(betterproto.Message):
    sub_msg: "TwinInt12Message" = betterproto.message_field(1)


@dataclass
class ParamsMessage(betterproto.Message):
    field_1: "Field1SubMessage" = betterproto.message_field(1)
    field_array_2: List["TwinIntMessage"] = betterproto.message_field(2)
    int_3: int = betterproto.uint32_field(3, group="int_2")
    sub_messsage_4: "Field4SubMessage" = betterproto.message_field(4)


@dataclass
class ShortsRequestPaginationSubSub(betterproto.Message):
    video_id: str = betterproto.string_field(4)
    uint_6: int = betterproto.uint32_field(6, group="int_6")
    uint_15: int = betterproto.uint32_field(15, group="int_15")
    uint_25: int = betterproto.uint32_field(25, group="int_25")


@dataclass
class ShortsRequestPaginationSub(betterproto.Message):
    sub_message_4: "ShortsRequestPaginationSubSub" = betterproto.message_field(4)
    uint_6: int = betterproto.int32_field(6, group="int_6")
    string_8: bytes = betterproto.bytes_field(8)


@dataclass
class ShortsRequestPagination(betterproto.Message):
    sub_message_1: "ShortsRequestPaginationSub" = betterproto.message_field(53)


@dataclass
class ShortsRequestSub(betterproto.Message):
    """Principal message for shorts"""

    sort_type: bytes = betterproto.bytes_field(2)
    pagination: bytes = betterproto.bytes_field(3)


@dataclass
class ShortContinuationToken(betterproto.Message):
    params: "ShortsRequestSub" = betterproto.message_field(80226972)



import json
from typing import List, Tuple

import dateparser
import requests

from youtube_simple_scraper.continuation_token import CommentsContinuationTokenBuilder
from youtube_simple_scraper.counters import comment_counter_to_int
from youtube_simple_scraper.list_videos_request import USER_AGENT


def get_shorts(channel_id: str, continuation_token: str = "") -> Tuple[List[dict], str]:
    url = "https://www.youtube.com/youtubei/v1/browse?prettyPrint=false"
    payload_dict = {
        "context": {
            "client": {
                "deviceModel": "",
                "userAgent": USER_AGENT,
                "clientName": "WEB",
                "clientVersion": "2.20240530.02.00",
                "osName": "Macintosh",
                "osVersion": "10_15_7",
                "originalUrl": "https://www.youtube.com/@IbaiLlanos/shorts",
            },
        }
    }
    if continuation_token == "":
        payload_dict["browseId"] = channel_id
        payload_dict["params"] = "EgZzaG9ydHPyBgUKA5oBAA%3D%3D"
    else:
        payload_dict["continuation"] = continuation_token

    payload = json.dumps(payload_dict)
    headers = {
        'accept': '*/*',
        'accept-language': 'en',
        'content-type': 'application/json',
        'origin': 'https://www.youtube.com',
        'priority': 'u=1, i',
        'sec-ch-ua': '"Google Chrome";v="125", "Chromium";v="125", "Not.A/Brand";v="24"',
        'sec-ch-ua-arch': '"arm"',
        'sec-ch-ua-bitness': '"64"',
        'user-agent': USER_AGENT,
    }

    response = requests.request("POST", url, headers=headers, data=payload)
    j = response.json()
    raw_videos = []
    next_continuation_token = ""
    # /onResponseReceivedActions/0/appendContinuationItemsAction/continuationItems/48/continuationItemRenderer/continuationEndpoint/continuationCommand/token
    # /onResponseReceivedActions/0/appendContinuationItemsAction/continuationItems/0/richItemRenderer/content/reelItemRenderer/headline/simpleText
    reel_item_renderers = []
    if "onResponseReceivedActions" in j:
        for r in j["onResponseReceivedActions"]:
            if 'appendContinuationItemsAction' not in r:
                continue
            for c in r["appendContinuationItemsAction"]["continuationItems"]:
                if 'continuationItemRenderer' in c:
                    next_continuation_token = \
                        c["continuationItemRenderer"]["continuationEndpoint"]["continuationCommand"]["token"]
                if 'richItemRenderer' not in c:
                    continue
                reel_item_renderers.append(c["richItemRenderer"]["content"]["reelItemRenderer"])
    elif "contents" in j and "twoColumnBrowseResultsRenderer" in j["contents"] and "tabs" in \
            j["contents"]["twoColumnBrowseResultsRenderer"]:
        for t in j["contents"]["twoColumnBrowseResultsRenderer"]["tabs"]:
            if ('tabRenderer' not in t or 'content' not in t["tabRenderer"]
                    or 'richGridRenderer' not in t["tabRenderer"]["content"]):
                continue
            for c in t["tabRenderer"]["content"]["richGridRenderer"]["contents"]:
                if 'continuationItemRenderer' in c:
                    next_continuation_token = c['continuationItemRenderer']['continuationEndpoint']['continuationCommand'][
                        'token']
                if "richItemRenderer" not in c:
                    continue
                reel_item_renderers.append(c["richItemRenderer"]["content"]["reelItemRenderer"])

    for r in reel_item_renderers:
        video_id = r["videoId"]
        title = r["headline"]["simpleText"]
        thumbnail = r['thumbnail']['thumbnails'][0]['url']
        view_count = r['viewCountText']['simpleText']
        raw_videos.append({
            "id": video_id,
            "title": title,
            "thumbnail": thumbnail,
            "view_count": comment_counter_to_int(view_count)
        })
    return raw_videos, next_continuation_token


def get_short_comments(short_id: str, continuation_token: str = "") -> Tuple[List[dict], str]:
    if continuation_token == "":
        continuation_token = CommentsContinuationTokenBuilder.build_start_short_comment_cont_token(short_id)

    url = "https://www.youtube.com/youtubei/v1/browse?prettyPrint=false"
    payload = json.dumps({
        "context": {
            "client": {
                "deviceModel": "",
                "clientName": "WEB",
                "clientVersion": "2.20240530.02.00",
                "osVersion": "10_15_7"
            }
        },
        "continuation": continuation_token
    })
    headers = {
        'accept': '*/*',
        'accept-language': 'en',
        'content-type': 'application/json',
        'user-agent': USER_AGENT,
    }
    response = requests.request("POST", url, headers=headers, data=payload)
    j = response.json()
    raw_comments = []
    continuation_token = ""
    if "contents" in j and "twoColumnBrowseResultsRenderer" in j["contents"] and "tabs" in \
            j["contents"]["twoColumnBrowseResultsRenderer"]:
        for t in j["contents"]["twoColumnBrowseResultsRenderer"]["tabs"]:
            if ('tabRenderer' not in t or 'content' not in t["tabRenderer"] or
                    'richGridRenderer' not in t["tabRenderer"]["content"]):
                continue
            for c in t["tabRenderer"]["content"]["richGridRenderer"]["contents"]:
                if "continuationItemRenderer" in c:
                    continuation_token = c['continuationItemRenderer']['continuationEndpoint'][
                        'continuationCommand']['token']

    if 'frameworkUpdates' not in j or 'entityBatchUpdate' not in j['frameworkUpdates']:
        return [], continuation_token
    for c in j["frameworkUpdates"]["entityBatchUpdate"]["mutations"]:
        if 'payload' not in c or 'commentEntityPayload' not in c['payload']:
            continue
        p = c['payload']['commentEntityPayload']
        if 'properties' not in p or 'content' not in p['properties']:
            continue
        content = p['properties']['content']['content']
        published_time = p['properties']['publishedTime']
        d = dateparser.parse(published_time)
        likes = 0
        if 'toolbar' in p and 'likeCountA11y' in p['toolbar']:
            likes = int(p['toolbar']['likeCountA11y'].split(" ")[0])
        reply_count = 0
        if 'toolbar' in p and 'replyCount' in p['toolbar'] and p['toolbar']['replyCount'] != "":
            reply_count = int(p['toolbar']['replyCount'])
        raw_comments.append({
            "id": p['properties']['commentId'],
            "text": content,
            "date": d,
            "likes": likes,
            "author": p['author']['displayName'],
            "reply_count": reply_count
        })
    return raw_comments, continuation_token


if __name__ == '__main__':
    cid = "UCaY_-ksFSQtTGk0y1HA_3YQ"
    v, token = get_shorts(cid)
    # while True:
    #     if token == "":
    #         break
    #     v2, token = get_shorts(cid, token)
    #     print(v2)
    #     v.extend(v2)

    comments, s_token = get_short_comments(v[0]["id"])
    print(comments)



import json
from datetime import datetime
from typing import List

from pydantic import BaseModel

from tests.utils import read_test_file, FILE_WITH_PAGE_HEADER_RENDERER, EXPECTED_VIDEOS
from youtube_simple_scraper.list_videos import ApiVideoListRepository, Video
import unittest


class VideoList(BaseModel):
    videos: List[Video]


class TestApiVideoListRepository(unittest.TestCase):

    def test_extract_videos_from_tab_renderer(self):
        file = read_test_file(FILE_WITH_PAGE_HEADER_RENDERER)
        repository = ApiVideoListRepository()
        file_dict = json.loads(file)
        videos = repository._extract_raw_videos(file_dict)
        exp_json = read_test_file(EXPECTED_VIDEOS)
        expected_videos = VideoList.model_validate_json(exp_json).videos
        self.assertIsNotNone(videos)
        self.assertEqual(len(videos), len(expected_videos))
        for i in range(len(videos)):
            self.assertIsNotNone(videos[i].date)
            videos[i].date = datetime.now()
            expected_videos[i].date = videos[i].date
            self.assertEqual(videos[i], expected_videos[i])
        self.assertEqual(videos, expected_videos)



import abc
from typing import List

from youtube_simple_scraper.entities import VideoComment
from youtube_simple_scraper.list_short import get_shorts, get_short_comments
from youtube_simple_scraper.list_videos_request import get_comments_from_api


class VideoCommentRepository(abc.ABC):
    @abc.abstractmethod
    def next(self, video_id: str) -> List[VideoComment]:
        pass


class ApiVideoCommentRepository(VideoCommentRepository):

    def __init__(self):
        self._tokens = {}

    def next(self, video_id: str) -> List[VideoComment]:
        raw_comments, next_token = get_comments_from_api(video_id, self._tokens.get(video_id, ""))
        self._tokens[video_id] = next_token
        comments: List[VideoComment] = []
        for comment in raw_comments:
            comments.append(VideoComment(
                id=comment["id"],
                text=comment["comment_text"],
                user=comment["author"],
                date=comment["published_time"],
                likes=comment["likes"]
            ))
        return comments


class ApiShortVideoCommentRepository(VideoCommentRepository):
    def __init__(self):
        self._tokens = {}

    def next(self, short_id: str) -> List[VideoComment]:
        token = ""
        if short_id in self._tokens:
            token = self._tokens[short_id]
        raw_comments, s_token = get_short_comments(short_id, token)
        comments: List[VideoComment] = []
        for comment in raw_comments:
            comments.append(VideoComment(
                id=comment["id"],
                text=comment["text"],
                user=comment["author"],
                date=comment["date"],
                likes=comment["like"]
            ))
        return comments



from setuptools import setup, find_packages


def read_version():
    with open('youtube_simple_scraper/__init__.py') as f:
        for line in f:
            if line.startswith('__version__'):
                return line.split('=')[1].strip().replace("'", '').replace('"', '')
    return '0.0.0'


def increment_and_update_version():
    version = read_version()
    major, minor, patch = map(int, version.split('.'))
    patch += 1
    new_version = f"{major}.{minor}.{patch}"

    with open('youtube_simple_scraper/__init__.py', 'r+') as f:
        content = f.read()
        f.seek(0)
        f.truncate()
        f.write(content.replace(version, new_version))

    return new_version.replace("'", '').replace('"', '')


setup(
    name='youtube_simple_scraper',
    version=increment_and_update_version(),
    packages=find_packages(include=['youtube_simple_scraper', 'youtube_simple_scraper.*']),
    include_package_data=True,
    install_requires=[
        'pydantic~=2.7.1',
        'requests~=2.31.0',
        'beautifulsoup4~=4.12.3',
        'python-dateutil~=2.9.0.post0',
        'dateparser~=1.2.0',
        'betterproto~=1.2.5',
    ],
    author='Hector Oliveros',
    author_email='hector.oliveros.leon@gmail.com',
    description='A simple scraper for Youtube',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Eitol/youtube_simple_scraper',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.12',
    license="MIT",
)



import json
import logging
from typing import List, Tuple

import dateparser
import requests

from youtube_simple_scraper.entities import GetChannelOptions, Video, Channel, VideoComment, VideoListRepository, \
    VideoType
from youtube_simple_scraper.list_short import get_shorts
from youtube_simple_scraper.list_video_comments import VideoCommentRepository

from youtube_simple_scraper.list_videos_request import build_list_videos_request_payload, \
    build_list_videos_request_headers, find_channel_basic_info

default_logger = logging.getLogger(__name__)


class ApiVideoListRepository(VideoListRepository):

    def __init__(self, video_comment_repo: VideoCommentRepository, shorts_comment_repo: VideoCommentRepository,
                 logger: logging.Logger = default_logger):
        self._video_comment_repo = video_comment_repo
        self._shorts_comment_repo = shorts_comment_repo
        self._logger = logger

    def get_channel(self, channel_name: str, opts: GetChannelOptions) -> Channel:
        self._logger.info(f"Fetching channel {channel_name}")
        response_body = find_channel_basic_info(channel_name)
        self._logger.info(f"Found channel {channel_name}")
        for attempt in range(0, 3):
            self._logger.info(f"Attempt {attempt + 1} to get channel {channel_name}")
            channel, videos_continuation_token, is_valid = self._extract_channel_basic_info(response_body)
            channel.name = channel_name
            if not is_valid:
                self._logger.info(f"Invalid response for channel {channel_name}")
                continue

            channel.videos = self._fetch_next_videos(channel, opts, videos_continuation_token)
            self._add_comments_to_videos(opts, channel.videos)

            channel.shorts = self._fetch_shorts_videos(channel, opts)
            self._add_comments_to_shorts(opts, channel.shorts)

            return channel

    def _fetch_next_videos(self, channel: Channel, opts: GetChannelOptions, continuation_token: str) -> List[Video]:
        initial_videos_added = False
        total_videos = []
        page_count = 0
        while True:
            if self._should_stop_fetch_videos(total_videos, page_count, opts):
                self._logger.info(f"Videos stop condition reached for channel {channel.name}")
                break
            self._logger.info(f"Fetching page {page_count} for channel {channel.name}")
            response_body = self._send_next_video_page_request(channel.name, channel.id, continuation_token)
            videos = self._extract_raw_videos(response_body)
            if not initial_videos_added:
                videos = channel.videos + videos
                channel.videos = []
                initial_videos_added = True
            continuation_token = self._extract_token(response_body)
            total_videos.extend(videos)
            page_count += 1
            if not continuation_token:
                self._logger.info(f"No more pages for channel {channel.name}")
                break
        return total_videos

    def _add_comments_to_videos(self, opts, videos):
        for i in range(len(videos)):
            video = videos[i]
            comments_page = 0
            while True:
                if self._should_stop_fetch_video_comments(video, video.comments, comments_page, opts):
                    self._logger.info(f"Stop condition reached for video {video.id}")
                    break
                self._logger.info(f"Fetching comments for video {video.url} page {comments_page + 1}")
                comments = self._video_comment_repo.next(video.id)
                if len(comments) == 0:
                    self._logger.info(f"No comments found for video {video.url}")
                    continue
                self._logger.info(f"Found {len(comments)} comments for video {video.url}")
                videos[i].comments.extend(comments)
                comments_page += 1

    def _fetch_shorts_videos(self, channel: Channel, opts: GetChannelOptions) -> List[Video]:
        total_videos = []
        page_count = 0
        token = ""
        while True:
            if self._should_stop_fetch_shorts(total_videos, page_count, opts):
                self._logger.info(f"Shorts stop condition reached for channel {channel.name}")
                break
            self._logger.info(f"Fetching shorts page {page_count} for channel {channel.name}")
            raw_shorts, token = get_shorts(channel.id, token)
            for short in raw_shorts:
                video = Video(
                    id=short["id"],
                    title=short["title"],
                    thumbnail_url=short["thumbnail"],
                    view_count=short["view_count"],
                    comments=[],
                    video_type=VideoType.SHORT
                )
                total_videos.append(video)
            page_count += 1
            if not token or len(raw_shorts) == 0:
                self._logger.info(f"No more shorts pages for channel {channel.name}")
                break
        return total_videos

    def _add_comments_to_shorts(self, opts, shorts: List[Video]) -> None:
        for i in range(len(shorts)):
            short = shorts[i]
            comments_page = 0
            while True:
                if self._should_stop_fetch_shorts_comments(short, short.comments, comments_page, opts):
                    self._logger.info(f"Stop condition reached for short {short.id}")
                    break
                self._logger.info(f"Fetching comments for short {short.url} page {comments_page + 1}")
                comments = self._video_comment_repo.next(short.id)
                if len(comments) == 0:
                    self._logger.info(f"No comments found for short {short.url}")
                    continue
                self._logger.info(f"Found {len(comments)} comments for short {short.url}")
                shorts[i].comments.extend(comments)
                comments_page += 1

    @staticmethod
    def _should_stop_fetch_videos(videos: List[Video], page: int, opts: GetChannelOptions) -> bool:
        return any([condition.should_stop(videos, page) for condition in opts.list_video_stop_conditions])

    @staticmethod
    def _should_stop_fetch_shorts(videos: List[Video], page: int, opts: GetChannelOptions) -> bool:
        return any([condition.should_stop(videos, page) for condition in opts.list_short_stop_conditions])

    @staticmethod
    def _should_stop_fetch_video_comments(video: Video, comments: List[VideoComment], page: int,
                                          opts: GetChannelOptions) -> bool:
        return any([
            condition.should_stop(video, comments, page) for condition in opts.list_video_comment_stop_conditions])

    @staticmethod
    def _should_stop_fetch_shorts_comments(video: Video, comments: List[VideoComment], page: int,
                                           opts: GetChannelOptions) -> bool:
        return any([
            condition.should_stop(video, comments, page) for condition in opts.list_short_comment_stop_conditions])

    @staticmethod
    def _send_next_video_page_request(channel_id: str, browse_id: str, continuation_token: str) -> dict:
        payload = build_list_videos_request_payload(channel_id, browse_id, continuation_token)
        headers = build_list_videos_request_headers(channel_id)
        payload_json = json.dumps(payload)
        url = "https://www.youtube.com/youtubei/v1/browse?prettyPrint=false"
        resp = requests.post(url, data=payload_json, headers=headers)
        return resp.json()

    @staticmethod
    def _abbreviate_number_to_number(number: str) -> int:
        if '.' in number:
            number = number.replace('.', '')
            if 'M' in number:
                number = number.replace('M', '00000')
            elif 'K' in number:
                number = number.replace('K', '00')
        else:
            if 'M' in number:
                number = number.replace('M', '000000')
            elif 'K' in number:
                number = number.replace('K', '000')
        try:
            subscriber_count = int(number.split(" ")[0].replace(",", ""))
        except:
            subscriber_count = 0
        return subscriber_count

    def _extract_channel_basic_info(self, response_body: dict) -> Tuple[Channel, str, bool]:
        if "header" not in response_body:
            return self._build_empty_channel(), "", False
        if "c4TabbedHeaderRenderer" in response_body["header"]:
            channel, token = self._extract_channel_from_c4_tabbed_header_renderer_tag(response_body)
        else:
            channel, token = self._extract_channel_from_page_header_renderer_tag(response_body)
        channel.videos = self._extract_raw_videos(response_body)
        return channel, token, True

    def _extract_raw_videos(self, response_body: dict) -> List[Video]:
        videos = self._extract_videos_from_next_videos_page_response(response_body)
        if len(videos) > 0:
            return self._adapt_raw_videos(videos)
        try:
            tabs = response_body["contents"]["twoColumnBrowseResultsRenderer"]["tabs"]
        except:
            return []

        for tab in tabs:
            try:
                tab_renderer = tab["tabRenderer"]
            except:
                return []
            if 'content' not in tab_renderer:
                continue
            raw_videos: List[dict] = self._extract_raw_videos_from_contents_tag(tab_renderer)
            return self._adapt_raw_videos(raw_videos)

    @classmethod
    def _build_empty_channel(cls):
        channel: Channel = Channel.construct(
            id="",
            title="",
            target_id="",
            description="",
            subscriber_count=0,
            video_count=0,
            url="",
            videos=[]
        )
        return channel

    @classmethod
    def _extract_videos_from_next_videos_page_response(cls, response_body: dict) -> List[dict]:
        if "onResponseReceivedActions" not in response_body:
            return []
        raw_videos = []
        for ra in response_body["onResponseReceivedActions"]:
            if 'appendContinuationItemsAction' not in ra:
                continue
            for ci in ra['appendContinuationItemsAction']['continuationItems']:
                if 'richItemRenderer' not in ci:
                    continue
                rir = ci['richItemRenderer']
                if 'content' not in rir:
                    continue
                content = rir['content']
                if 'videoRenderer' not in content:
                    continue
                raw_videos.append(content['videoRenderer'])
        return raw_videos

    @classmethod
    def _extract_channel_from_page_header_renderer_tag(cls, response_body: dict) -> Tuple[Channel, str]:
        channel = cls._build_empty_channel()
        t = response_body["header"]["pageHeaderRenderer"]
        try:
            channel.target_id = t["content"]["pageHeaderViewModel"]["attribution"]["attributionViewModel"][
                "suffix"]["commandRuns"][0]["onTap"]["innertubeCommand"][
                "showEngagementPanelEndpoint"]["engagementPanel"]["engagementPanelSectionListRenderer"][
                "content"]["sectionListRenderer"][
                "contents"][0]["itemSectionRenderer"]["targetId"]
        except:
            pass
        rows = t['content']['pageHeaderViewModel']['metadata']['contentMetadataViewModel']['metadataRows']
        for row in rows:
            for part in row['metadataParts']:
                if 'content' in part:
                    content = part['content']
                elif 'text' in part:
                    content = part['text']['content']
                else:
                    continue
                if 'subscriber' in content:
                    subscriber_count_text = content.split(' ')[0]
                    subscriber_count_text = subscriber_count_text.split(' ')[0]
                    channel.subscriber_count = cls._abbreviate_number_to_number(subscriber_count_text)
                elif 'video' in content:
                    video_count_text = content.split(' ')[0]
                    channel.video_count = cls._abbreviate_number_to_number(video_count_text)
        m = response_body["metadata"]["channelMetadataRenderer"]
        channel.id = m['externalId']
        channel.title = m['title']
        channel.description = m['description']
        token = cls._extract_token(response_body)
        return channel, token

    @classmethod
    def _extract_channel_from_c4_tabbed_header_renderer_tag(cls, response_body: dict) -> Tuple[Channel, str]:
        channel = cls._build_empty_channel()
        try:
            channel.target_id = response_body["contents"]["twoColumnBrowseResultsRenderer"]["tabs"][1]["tabRenderer"][
                "content"]["richGridRenderer"]["header"]["feedFilterChipBarRenderer"]["contents"][0][
                "chipCloudChipRenderer"]["navigationEndpoint"]["continuationCommand"]["command"][
                "showReloadUiCommand"]["targetId"]
        except:
            pass
        t = response_body["header"]["c4TabbedHeaderRenderer"]
        channel.video_count = cls._abbreviate_number_to_number(t['videosCountText']['runs'][0]['text'])
        channel.id = t['channelId']
        channel.title = t['title']
        subscriber_count_text = t['subscriberCountText']['simpleText']
        channel.subscriber_count = cls._abbreviate_number_to_number(subscriber_count_text)
        channel.description = t['tagline']["channelTaglineRenderer"]["content"]
        token = cls._extract_token(response_body)
        return channel, token

    @classmethod
    def _extract_raw_videos_from_contents_tag(cls, tab_renderer) -> List[dict]:
        raw_videos: List[dict] = []
        if "sectionListRenderer" in tab_renderer["content"]:
            contents = tab_renderer["content"]["sectionListRenderer"]["contents"]
        else:
            contents = tab_renderer["content"]["richGridRenderer"]["contents"]
        for content in contents:
            if "itemSectionRenderer" in content:
                c = content["itemSectionRenderer"]["contents"][0]["shelfRenderer"]["content"]
            else:
                c = content
            if "horizontalListRenderer" in c:
                hlr = c["horizontalListRenderer"]
                for item in hlr["items"]:
                    raw_videos.append(item['gridVideoRenderer'])
            elif "expandedShelfContentsRenderer" in c:
                hlr = c["expandedShelfContentsRenderer"]
                for item in hlr["items"]:
                    raw_videos.append(item['videoRenderer'])
            elif "richItemRenderer" in c:
                raw_videos.append(c["richItemRenderer"]["content"]["videoRenderer"])
            else:
                continue
        return raw_videos

    @staticmethod
    def _extract_token(response_body: dict) -> str:
        try:
            tabs = response_body["contents"]["twoColumnBrowseResultsRenderer"]["tabs"]
            for tab in tabs:
                if "content" not in tab["tabRenderer"]:
                    continue
                contents = tab["tabRenderer"]["content"]["richGridRenderer"]["contents"]
                for i in range(len(contents) - 1, -1, -1):
                    if "continuationItemRenderer" not in contents[i] or "continuationEndpoint" not in \
                            contents[i]["continuationItemRenderer"]:
                        continue
                    return contents[i]["continuationItemRenderer"]["continuationEndpoint"][
                        "continuationCommand"]["token"]
        except:
            pass
        try:
            if "onResponseReceivedActions" in response_body:
                for action in response_body["onResponseReceivedActions"]:
                    if "appendContinuationItemsAction" in action:
                        for item in action["appendContinuationItemsAction"]["continuationItems"]:
                            if "continuationItemRenderer" in item:
                                if "continuationEndpoint" in item["continuationItemRenderer"]:
                                    return item["continuationItemRenderer"]["continuationEndpoint"][
                                        "continuationCommand"]["token"]
        except:
            pass

        try:
            return response_body["header"]["pageHeaderRenderer"]["content"]["pageHeaderViewModel"][
                "description"]["descriptionPreviewViewModel"]["rendererContext"]["commandContext"][
                "onTap"]["innertubeCommand"]["showEngagementPanelEndpoint"]["engagementPanel"][
                "engagementPanelSectionListRenderer"]["content"]["sectionListRenderer"][
                "contents"][0]["itemSectionRenderer"]["contents"][0]["continuationItemRenderer"][
                "continuationEndpoint"]["continuationCommand"]["token"]
        except:
            pass
        try:
            return response_body["header"]["pageHeaderRenderer"]["content"]["pageHeaderViewModel"][
                "attribution"]["attributionViewModel"][
                "suffix"]["commandRuns"][0]["onTap"]["innertubeCommand"]["showEngagementPanelEndpoint"][
                "engagementPanel"]["engagementPanelSectionListRenderer"]["content"]["sectionListRenderer"]["contents"][
                0]["itemSectionRenderer"]["contents"][0]["continuationItemRenderer"]["continuationEndpoint"][
                "continuationCommand"]["token"]
        except:
            pass
        try:
            return response_body['header']['c4TabbedHeaderRenderer']['tagline']['channelTaglineRenderer'][
                'moreEndpoint']['showEngagementPanelEndpoint'][
                'engagementPanel']['engagementPanelSectionListRenderer']['content'][
                'sectionListRenderer']['contents'][0]['itemSectionRenderer']['contents'][0][
                'continuationItemRenderer']['continuationEndpoint']['continuationCommand']['token']
        except:
            pass
        return ""

    def _adapt_raw_videos(self, raw_videos: List[dict]) -> List[Video]:
        extracted_videos: List[Video] = []
        for raw_video in raw_videos:

            view_count_text = raw_video.get("viewCountText", {}).get("simpleText", '')
            view_count = 0
            if view_count_text:
                view_count = int(view_count_text.split(' ')[0].replace(",", ""))
            else:
                pass
            human_date = raw_video.get("publishedTimeText", {}).get("simpleText", "")
            date = dateparser.parse(human_date, languages=["en"])
            if date is None:
                pass
            title = ''
            if "title" in raw_video:
                if "simpleText" in raw_video["title"]:
                    title = raw_video["title"]["simpleText"]
                elif "runs" in raw_video["title"]:
                    title = raw_video["title"]["runs"][0]["text"]
            description = raw_video.get("description", {}).get("simpleText", "")
            if not description:
                try:
                    description = raw_video['descriptionSnippet']['runs'][0]['text']
                except:
                    pass
            thumbnail_url = raw_video.get("thumbnail", {}).get("thumbnails", [{}])[0].get("url", "")
            if not thumbnail_url:
                try:
                    thumbnail_url = \
                        raw_video['richThumbnail']['movingThumbnailRenderer']['movingThumbnailDetails']['thumbnails'][0]
                except:
                    pass
            likes = 0
            likes_str = (
                raw_video.get("factoid", {}).get("factoidRenderer", {})
                .get("factoidRenderer", {}).get("values", {}).get(0, {}).get("simpleText", ""))
            if likes_str:
                likes = int(likes_str.split(' ')[0].replace(",", ""))
            if likes == 0:
                likes = int(raw_video.get("likes", 0))

            v = Video(
                id=raw_video["videoId"],
                title=title,
                description=description,
                date=date,
                view_count=view_count,
                like_count=likes,
                comment_count=int(raw_video.get("commentCount", 0)),
                thumbnail_url=thumbnail_url,
                comments=[],
                video_type=VideoType.VIDEO
            )
            extracted_videos.append(v)
        return self._sort_videos_by_date(extracted_videos)

    @staticmethod
    def _sort_videos_by_date(video_list: List[Video]) -> List[Video]:
        return sorted(video_list, key=lambda x: x.date, reverse=True)

    @classmethod
    def _adapt_raw_comments(cls, raw_comments: dict):
        pass
